from abc import ABCMeta, abstractmethod
try:
    from flask import (
        Flask as flask_app,
        has_request_context as flask_has_request_context,
        session as flask_session,
        request as flask_request,
        redirect as flask_redirect,
        g as flask_g,
        url_for as flask_url_for
        )
    from .flask_blueprint import FlaskAADEndpoints # this is where our auth-related endpoints are defined
except:
    pass

from .context import IdentityContextData
from typing import Any
from functools import wraps

# decorator to make sure access within request context
def require_request_context(f):
    @wraps(f)
    def assert_context(self, *args, **kwargs):
        if not flask_has_request_context():
            self.logger.info(f"{self.__class__.__name__}.{f.__name__}: No request context, aborting")
        else:
            return f(self, *args, **kwargs)
    return assert_context

class IdentityWebContextAdapter(metaclass=ABCMeta):
    """Context Adapter abstract base class. Extend this to enable IdentityWebPython to
    work within any environment (e.g. Flask, Django, Windows Desktop app, etc) """
    @abstractmethod
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def _on_request_init(self) -> None: 
        pass

    @abstractmethod
    def _on_request_end(self) -> None:
        pass

    # TODO: make this dictionary key name configurable on app init
    @abstractmethod
    def attach_identity_web_util(self, identity_web: 'IdentityWebPython') -> None:
        pass

    @abstractmethod # @property
    def has_context(self) -> bool:
        pass
    
    @abstractmethod # @property
    @require_request_context
    def identity_context_data(self) -> 'IdentityContextData':
        pass
    

    @abstractmethod # @property
    @require_request_context
    def session(self) -> None:
        # TODO: set session attr so can concrete implement here?
        pass

    @abstractmethod
    @require_request_context
    def clear_session(self) -> None:
        # TODO: clear ONLY msidweb session stuff
        # TODO: concrete implement here instead?
        pass

    @require_request_context
    def get_value_from_session(self, key: str, default: Any = None) -> Any:
        return self.session.get(key, default)

    
    @require_request_context
    def get_request_param(self, key: str, default: Any = None) -> Any:
        return self._get_request_params_as_dict(key, default)

    @abstractmethod
    @require_request_context
    def redirect_to_absolute_url(self, absolute_url: str) -> None:
        #TODO: set attr redirect on init, so concrete method can be used for all frmwrks
        pass

    @abstractmethod
    @require_request_context
    def get_request_params_as_dict(self) -> dict:
        pass

    @abstractmethod
    @require_request_context
    def _deserialize_identity_context_data_from_session(self) -> 'IdentityContextData':
        pass

    @abstractmethod
    @require_request_context
    def _serialize_identity_context_data_to_session(self) -> None:
        pass

class FlaskContextAdapter(IdentityWebContextAdapter):
    """Context Adapter to enable IdentityWebPython to work within the Flask environment"""
    def __init__(self, app) -> None:
        assert isinstance(app, flask_app)
        super().__init__()
        self.app = app
        with self.app.app_context():
            self.logger = app.logger
            app.before_request(self._on_request_init)
            app.after_request(self._on_request_end)

    @property
    @require_request_context
    def identity_context_data(self) -> 'IdentityContextData':
        # TODO: make the key name configurable
        self.logger.debug("Getting identity_context from g")
        identity_context_data = flask_g.get(IdentityContextData.SESSION_KEY)
        if not identity_context_data:
            identity_context_data = self._deserialize_identity_context_data_from_session()
            setattr(flask_g, IdentityContextData.SESSION_KEY, identity_context_data)
        return identity_context_data

    # method is called when flask gets an app/request context
    # Flask-specific startup here?    
    def _on_request_init(self) -> None:
        try:
            idx = self.identity_context_data # initialize it so it is available to request context
        except Exception as ex:
            self.logger.error(f'Adapter failed @ _on_request_init\n{ex}')

    # this is for saving any changes to the identity_context_data
    def _on_request_end(self, response_to_return=None) -> None:
        try:
            if IdentityContextData.SESSION_KEY in flask_g:
                self._serialize_identity_context_data_to_session()
        except Exception as ex:
            self.logger.error(f'flask adapter failed @ _on_request_ended\n{ex}')

        return response_to_return

    # TODO: order is reveresed? create id web first, then attach flask adapter to it!?
    def attach_identity_web_util(self, identity_web: 'IdentityWebPython') -> None:
        """attach the identity web instance to session so it is accessible everywhere.
        e.g., ms_id_web = current_app.config.get("ms_identity_web")\n
        Also attaches the application logger."""
        aad_config = identity_web.aad_config
        config_key = aad_config.flask.id_web_configs

        with self.app.app_context():
            self.app.config[config_key] = aad_config

        identity_web.set_logger(self.logger)
        auth_endpoints = FlaskAADEndpoints(identity_web)
        self.app.context_processor(lambda: dict(ms_id_url_for=auth_endpoints.url_for))
        self.app.register_blueprint(auth_endpoints)        

    @property
    def has_context(self) -> bool:
        return flask_has_request_context()

    ### not sure if this method needs to be public yet :/
    @property
    @require_request_context
    def session(self) -> None:
        return flask_session

    # TODO: only clear IdWebPy vars
    @require_request_context
    def clear_session(self) -> None:
        """this function clears the session and refreshes context. TODO: only clear IdWebPy vars"""
        self.identity_context_data.clear()

    @require_request_context
    def redirect_to_absolute_url(self, absolute_url: str) -> None:
        """this function redirects to an absolute url"""
        return flask_redirect(absolute_url)

    @require_request_context
    def get_request_params_as_dict(self) -> dict:
        """this function returns the params dict from any flask request"""
        try:
            # this is query and form-post params merged,
            # preferring query param if there is a key collision
            return flask_request.values
        except:
            if self.logger is not None:
                self.logger.warning("failed to get param dict from request, substituting empty dict instead")
            return dict()

    # does this need to be public method?
    @require_request_context
    def _deserialize_identity_context_data_from_session(self) -> 'IdentityContextData':
        new_id_context_data = IdentityContextData()
        try:
            id_context_from_session = self.session.get(IdentityContextData.SESSION_KEY, dict())
            new_id_context_data.__dict__.update(id_context_from_session)
        except Exception as exception:
            self.logger.warning(f"failed to deserialize identity context from session: creating empty one\n{exception}")
        return new_id_context_data

    # does this need to be public method?
    @require_request_context
    def _serialize_identity_context_data_to_session(self) -> None:
        try:
            identity_context = self.identity_context_data
            if identity_context.has_changed:
                identity_context.has_changed = False
                identity_context = identity_context.__dict__
                self.session[IdentityContextData.SESSION_KEY] = identity_context
        except Exception as exception:
            self.logger.error(f"failed to serialize identity context to session.\n{exception}")

    






# the following class is incomplete
class DjangoContextAdapter(object):
    """Context Adapter to enable IdentityWebPython to work within the Django environment"""
    def __init__(self):
        raise NotImplementedError("not yet implemented")

    # method is called when getting app/request context
    def _on_context_init(self) -> None:
        self._has_context = True
    
    def _on_context_teardown(self, exception) -> None: 
        self._has_context = False
        if self.identity_context_data.has_changed:
            self.identity_context_data._save_to_session()

    # this function returns Django's request params.
    @require_request_context
    def get_request_params_as_dict(self, request: 'request' = None) -> dict:
        try:
            if request.method == "GET":
                return request.GET.dict()
            elif request.method == "POST" :
                return request.POST.dict()
            else:
                raise ValueError("Django request must be POST or GET")
        except:
            if self.logger is not None:
                self.logger.warning("Failed to get param dict, substituting empty dict instead")
            return dict()


from msal import ConfidentialClientApplication, SerializableTokenCache

from uuid import uuid4
from logging import Logger
from typing import Any
from functools import wraps
from .context import IdentityContextData
from .constants import *
from .adapters import IdentityWebContextAdapter
from .errors import *

# TODO: 
#  ##### IMPORTANT #####
# features:
# - do configurations work on multi-threaded flask environment? if not, attach them to current_app. configurations aren't stateful so this may be a moot point?
# - edit profile interaction required error on edit profile if no token_cache or expired?
# - password reset should use login hint/no interaction?
# - decorator for filter by security groups
# - decorator for app roles RBAC
# - auth failure handler to handle un-auth access?
# - implement more client_type options (single and multi tenant)
# - define django adapter: factor common adapter methods to a parent class that flask and django adapters inherit
#
# code quality:
# - more try catch blocks around sensitive failure-prone methods for gracefule error-handling
# 
# ###### LOWER PRIORITY ##########
# - remove any print statements
# - replace is comparators with == ?
# - cleanup / refactor constants file

def require_context_adapter(f):
    @wraps(f)
    def assert_adapter(self, *args, **kwargs):
        if not isinstance(self._adapter, IdentityWebContextAdapter) or not self._adapter.has_context:
            if self._logger:
                self._logger.info(f"{self.__class__.__name__}.{f.__name__}: invalid adapter or no request context, aborting")
            else:
                print(f"{self.__class__.__name__}.{f.__name__}: invalid adapter or no request context, aborting")
        return f(self, *args, **kwargs)
    return assert_adapter
        
class IdentityWebPython(object):

    def __init__(self, aad_config: 'AADConfig', adapter: IdentityWebContextAdapter = None, logger: Logger = None) -> None:
        self._logger = logger or Logger('IdentityWebPython')
        self._adapter = None
        self.aad_config = aad_config
        if adapter is not None:
             self.set_adapter(adapter)

    @property
    @require_context_adapter
    def id_data(self) -> IdentityContextData:
        return self._adapter.identity_context_data
    
    # TODO: make the call from the adapter to this and reverse the config process?
    def set_adapter(self, adapter: IdentityWebContextAdapter) -> None:                
        # if isinstance(adapter, FlaskContextAdapter):
        self._adapter = adapter
        adapter.attach_identity_web_util(self)
        # else:
        #     raise NotImplementedError(f"Currently, only the following adapters are supoprted: FlaskContextAdapter")
        
    def set_logger(self, logger: Logger) -> None:
        self._logger = logger

    def _client_factory(self, token_cache: SerializableTokenCache = None, b2c_policy: str = None, **msal_client_kwargs) -> ConfidentialClientApplication:
        client_config = self.aad_config.client.__dict__.copy() # need to make a copy since contents must be mutated
        client_config['authority'] = f'{self.aad_config.client.authority}{b2c_policy or ""}'
        if token_cache:
            client_config['token_cache'] = token_cache
        client_config.update(**msal_client_kwargs)

        return ConfidentialClientApplication(**client_config)        

    @require_context_adapter
    def get_auth_url(self, redirect_uri:str = None, b2c_policy: str = None, **msal_auth_url_kwargs):
        """ Gets the auth URL that the user must be redirected to. Automatically
            configures B2C if app type is set to B2C."""
        auth_req_options = self.aad_config.auth_request.__dict__.copy()
        auth_req_options.update(**msal_auth_url_kwargs)
        if redirect_uri:
            auth_req_options['redirect_uri'] = redirect_uri
        self._generate_and_append_state_to_context_and_request(auth_req_options)

        if self.id_data.authenticated:
            auth_req_options['login_hint'] = self.id_data._id_token_claims.get('preferred_username', None)

        if self.aad_config.type.authority_type == str(AuthorityType.B2C):
            if not b2c_policy:
                b2c_policy = self.aad_config.b2c.susi
            self._adapter.identity_context_data.last_used_b2c_policy = b2c_policy
            return self._client_factory(b2c_policy=b2c_policy).get_authorization_request_url(**auth_req_options)

        return self._client_factory().get_authorization_request_url(**auth_req_options)

    @require_context_adapter
    def process_auth_redirect(self, redirect_uri: str = None, response_type: str = None, afterwards_go_to_url: str = None) -> Any:
        req_params = self._adapter.get_request_params_as_dict() # grab the incoming request params
        try:
            # CSRF protection: make sure to check that state matches the one placed in the session in the previous step.
            # This check ensures this app + this same user session made the /authorize request that resulted in this redirect
            # This should always be the first thing verified on redirect.
            self._verify_state(req_params)
            
            self._logger.info("process_auth_redirect: state matches. continuing.")
            self._parse_redirect_errors(req_params)
            self._logger.info("process_auth_redirect: no errors found in request params. continuing.")
            
            # get the response_type that was requested, and extract the payload:
            resp_type = response_type or self.aad_config.auth_request.response_type or str(ResponseType.CODE)
            payload = self._extract_auth_response_payload(req_params, resp_type)
            cache = self._adapter.identity_context_data.token_cache
            redirect_uri = redirect_uri or self.aad_config.auth_request.redirect_uri or None

            if resp_type == str(ResponseType.CODE): # code request is default for msal-python if there is no response type specified
                # we should have a code. Now we must exchange the code for tokens.
                result = self._x_change_auth_code_for_token(payload, cache, redirect_uri)
            else:
                raise NotImplementedError(f"response_type {resp_type} is not yet implemented by ms_identity_web_python")
            self._process_result(result, cache)
            # self._verify_nonce() # one of the last steps TODO - is this required? msal python takes care of it?
        except AuthSecurityError as ase:
            self.remove_user()
            self._logger.error(f"process_auth_redirect: security violation {ase.args}")
            raise ase
        except OtherAuthError as oae:
            self.remove_user()
            self._logger.error(f"process_auth_redirect: other auth error {oae.args}")
            raise oae
        except B2CPasswordError as b2cpwe:
            self.remove_user()
            self._logger.error(f"process_auth_redirect: b2c pwd {b2cpwe.args}")
            pw_reset_url = self.get_auth_url(redirect_uri=redirect_uri, b2c_policy = self.aad_config.b2c.password)
            return self._adapter.redirect_to_absolute_url(pw_reset_url)
            # don't raise
        except TokenExchangeError as ter:
            self.remove_user()
            self._logger.error(f"process_auth_redirect: token xchange {ter.args}")
            raise ter
        except BaseException as other:
            self.remove_user()
            self._logger.error(f"process_auth_redirect: unknown error{other.args}")
            raise other
        
        #TODO: GET /auth/redirect?error=interaction_required&error_description=AADB2C90077%3a+User+does+not+have+an+existing+session+and+request+prompt+parameter+has+a+value+of+%27None%27.
        self._logger.info("process_auth_redirect: exiting auth code method. redirecting... ")
        return self._adapter.redirect_to_absolute_url(afterwards_go_to_url)

    @require_context_adapter
    def _x_change_auth_code_for_token(self, code: str, token_cache: SerializableTokenCache = None, redirect_uri = None) -> dict:
        # use the same policy that got us here: depending on /authorize request initiation
        id_context = self._adapter.identity_context_data
        if self.aad_config.type.authority_type == str(AuthorityType.B2C):
            b2c_policy = id_context.last_used_b2c_policy or self.aad_config.b2c.susi
            client = self._client_factory(token_cache=token_cache, b2c_policy=b2c_policy)
        else:
            client = self._client_factory(token_cache=token_cache)

        result = client.acquire_token_by_authorization_code(code, 
                                                   self.aad_config.auth_request.scopes,
                                                   redirect_uri,
                                                   id_context.nonce)
        return result

    @require_context_adapter
    def acquire_token_silently(self, scopes=None, account=None, authority=None, token_cache=None, **kwargs):
        # the params take precedence over settings file.
        id_data = self.id_data
        token_cache = token_cache or id_data.token_cache
        client = self._client_factory(token_cache=token_cache)

        silent_opts = dict()
        silent_opts.update(**kwargs)
        silent_opts['scopes'] = scopes or self.aad_config.auth_request.scopes
        silent_opts['account'] = account or client.get_accounts()[0]

        result = client.acquire_token_silent_with_error(**silent_opts)

        self._process_result(result, token_cache)

    @require_context_adapter
    def _process_result(self, result: dict, token_cache: SerializableTokenCache) -> None:
        if "error" not in result:
            self._logger.debug("process result: successful token response result!")
            # now we will place the token(s) and auth status into the context for later use:
            # self._logger.debug(json.dumps(result, indent=4, sort_keys=True))
            id_context = self._adapter.identity_context_data
            id_context.authenticated = True
            if 'id_token_claims' in result:
                id_context._id_token_claims = result['id_token_claims'] # TODO: if this is to stay in ctxt, use proper getter/setter
                id_context.username = id_context._id_token_claims.get('name', 'anonymous')
            if 'access_token' in result:
                id_context._access_token = result['access_token']
            id_context.has_changed = True                                   #TODO: update id_context to automatically do this when _id_token and accesstoken is assigned!!!!
            id_context.token_cache = token_cache
        else:
            raise TokenExchangeError("_process_result: auth failed: token request resulted in error\n"
                                        f"{result['error']}: {result.get('error_description', None)}")

    def _parse_redirect_errors(self, req_params: dict) -> None:
        # TODO implement all errors which affect program behaviour
        # https://docs.microsoft.com/en-us/azure/active-directory/develop/v2-oauth2-auth-code-flow
        if str(AADErrorResponse.ERROR_CODE_PARAM_KEY) in req_params:
            # we have an error. get the error code to interpret it:
            error_code = req_params.get(str(AADErrorResponse.ERROR_CODE_PARAM_KEY), None)
            if error_code.startswith(str(AADErrorResponse.B2C_FORGOT_PASSWORD_ERROR_CODE)):
                # it's a b2c password reset error
                raise B2CPasswordError("B2C password reset request")
            else:
                # ??? TODO: add more error types
                raise OtherAuthError("Unknown error while parsing redirect")

    def _extract_auth_response_payload(self, req_params: dict, expected_response_type: str) -> str:
        if expected_response_type == str(ResponseType.CODE):
            # if no response type in config, default response type of 'code' will have been assumed.
            return req_params.get(str(ResponseType.CODE), None)
        else:
            raise NotImplementedError("Only 'code' response is currently supported by identity utils")

    @require_context_adapter
    def sign_out(self, post_sign_out_url:str = None, username: str = None) -> Any:
        authority_type = self.aad_config.type.authority_type
        sign_out_url = self.aad_config.client.authority
        if authority_type == str(AuthorityType.B2C):
            sign_out_url = f'{sign_out_url}{self.aad_config.b2c.susi}{SignOut.ENDPOINT.value}'
        else:
            sign_out_url = f'{sign_out_url}{SignOut.ENDPOINT.value}'

        if post_sign_out_url:
            sign_out_url = f'{sign_out_url}?{SignOut.REDIRECT_PARAM_KEY.value}={post_sign_out_url}'
        return self._adapter.redirect_to_absolute_url(sign_out_url)
    
    @require_context_adapter
    def remove_user(self, username: str = None) -> None: #TODO: complete this so it doesn't just clear the session but removes user
        self._adapter.clear_session()
        # TODO e.g. if active username in id_context_'s username is not anonymous, remove it
        # remove id token
        # remote AT
        # remove token_cache
        # TODO: set auth_state_changed flag here
    
    @require_context_adapter
    def _generate_and_append_state_to_context_and_request(self, req_param_dict: dict) -> str:
        state = str(uuid4())
        req_param_dict[RequestParameter.STATE.value] = state
        self._adapter.identity_context_data.state = state
        return state
    
    @require_context_adapter
    def _verify_state(self, req_params: dict) -> None:
        state = req_params.get('state', None)
        session_state = self._adapter.identity_context_data.state
        # don't allow re-use of state
        self._adapter.identity_context_data.state = None
        # reject states that don't match
        if state is None or session_state != state:
            raise AuthSecurityError("Failed to match request state with session state")
    
    @require_context_adapter
    def _generate_and_append_nonce_to_context_and_request(self, req_param_dict: dict) -> str:
        nonce = str(uuid4())
        req_param_dict[RequestParameter.NONCE.value] = nonce
        self._adapter.identity_context_data.nonce = nonce
        return nonce

    @require_context_adapter
    def _verify_nonce(self, req_params: dict) -> None:
        nonce = req_params.get('nonce', None)
        session_nonce = self._adapter.identity_context_data.nonce
        # don't allow re-use of nonce
        self._adapter.identity_context_data.nonce = None
        # reject nonces that don't match
        if nonce is None or session_nonce != nonce:
            raise AuthSecurityError("Failed to match ID token nonce with session nonce")

    # TODO: enforce ID token expiry.
    # @decorator to ensure the user is authenticated
    # wrap this around your route    
    def login_required(self,f):
        @wraps(f)
        def assert_login(*args, **kwargs):
            if not self._adapter.identity_context_data.authenticated:
                raise NotAuthenticatedError
            # TODO: check if ID token is expired
            # if it is, take user to get re-authenticated.
            # TODO: upon returning from re-auth, user should get back to
            # where they were trying to go.
            return f(*args, **kwargs)
        return assert_login


try:    
    from ms_identity_web.errors import NotAuthenticatedError
    from django.conf import settings
    from django.shortcuts import render
except:
    pass

from .adapter import DjangoContextAdapter

ms_identity_web = settings.MS_IDENTITY_WEB

class MsalMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.
        self.ms_identity_web = ms_identity_web
    
    def process_exception(self, request, exception):
        if isinstance(exception, NotAuthenticatedError):
            if hasattr(settings, 'ERROR_TEMPLATE'):
                return render(request, settings.ERROR_TEMPLATE.format(exception.code))                
        return None

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        django_context_adapter = DjangoContextAdapter(request)
        self.ms_identity_web.set_adapter(django_context_adapter)
        django_context_adapter._on_request_init()
        
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        django_context_adapter._on_request_end()

        return response


try:
    from ms_identity_web import IdentityWebPython
    from ms_identity_web.context import IdentityContextData
    from ms_identity_web.adapters import IdentityWebContextAdapter
    from django.http.request import HttpRequest as DjangoHttpRequest
    from django.shortcuts import redirect as django_redirect
    import logging
except:
    pass

class DjangoContextAdapter(IdentityWebContextAdapter):
    """Context Adapter to enable IdentityWebPython to work within the Django environment"""
    def __init__(self, request: 'DjangoHttpRequest') -> None:
        # TODO: remove the following and add a middleware loaded before this one for global request/session context?
        self.request = request
        self._session = request.session
        self.logger = logging.getLogger('MsalMiddleWareLogger')

    @property
    def identity_context_data(self) -> 'IdentityContextData':
        # TODO: make the key name configurable
        self.logger.debug("Getting identity_context from request/session")
        identity_context_data = getattr(self.request, IdentityContextData.SESSION_KEY, None)
        if not identity_context_data:
            identity_context_data = self._deserialize_identity_context_data_from_session()
            setattr(self.request, IdentityContextData.SESSION_KEY, identity_context_data)
        return identity_context_data

    def _on_request_init(self) -> None:
        try:
            idx = self.identity_context_data # initialize it so it is available to request context
        except Exception as ex:
            self.logger.error(f'MsalMiddleware failed @ _on_request_init\n{ex}')

    # this is for saving any changes to the identity_context_data
    def _on_request_end(self) -> None:
        try:
            if getattr(self.request, IdentityContextData.SESSION_KEY, None):
                self._serialize_identity_context_data_to_session()
        except Exception as ex:
            self.logger.error(f'MsalMiddleware failed @ _on_request_ended\n{ex}')

    # TODO: order is reveresed? create id web first, then attach django adapter to it!?
    def attach_identity_web_util(self, identity_web: 'IdentityWebPython') -> None:
        """attach the identity web instance somewhere so it is accessible everywhere.
        e.g., ms_identity_web = current_app.config.get("ms_identity_web")\n
        Also attaches the application logger."""
        aad_config = identity_web.aad_config
        config_key = aad_config.django.id_web_configs

        setattr(self.request, config_key, aad_config)
        
    @property
    def has_context(self) -> bool:
        return True # TODO: remove this? not relevant for django?

    @property
    def session(self) -> None:
        return self._session

    # TODO: only clear IdWebPy vars
    def clear_session(self) -> None:
        """this function clears the session and refreshes context. TODO: only clear IdWebPy vars"""
        # TODO: clear ONLY msidweb session stuff
        self.session.flush()

    def redirect_to_absolute_url(self, absolute_url: str) -> None:
        """this function redirects to an absolute url"""
        return django_redirect(absolute_url)
        
    def get_request_params_as_dict(self) -> dict:
        try:
            if self.request.method == "GET":
                return self.request.GET.dict()
            elif self.request.method == "POST" :
                return self.request.POST.dict()
            else:
                raise ValueError("Django request must be POST or GET")
        except:
            if self.logger is not None:
                self.logger.warning("Failed to get param dict, substituting empty dict instead")
            return dict()

    # does this need to be public method?
    def _deserialize_identity_context_data_from_session(self) -> 'IdentityContextData':
        blank_id_context_data = IdentityContextData()
        try:
            id_context_from_session = self.session.get(IdentityContextData.SESSION_KEY, dict())
            blank_id_context_data.__dict__.update(id_context_from_session)
        except Exception as exception:
            self.logger.warning(f"failed to deserialize identity context from session: creating empty one\n{exception}")
        return blank_id_context_data

    # does this need to be public method?
    def _serialize_identity_context_data_to_session(self) -> None:
        try:
            identity_context = self.identity_context_data
            if identity_context.has_changed:
                identity_context.has_changed = False
                identity_context = identity_context.__dict__
                self.session[IdentityContextData.SESSION_KEY] = identity_context
        except Exception as exception:
            self.logger.error(f"failed to serialize identity context to session.\n{exception}")


from configparser import ConfigParser
import os
from .constants import AuthorityType, ClientType
from types import SimpleNamespace

class AADConfig(SimpleNamespace): # faster access to attributes with slots.
    @staticmethod
    def parse_json(file_path: str):
        import json
        from types import SimpleNamespace
        with open(file_path, 'r') as cfg:
            parsed_config = json.load(cfg, object_hook=lambda d: SimpleNamespace(**d))
        AADConfig.sanity_check_configs(parsed_config)
        return parsed_config

    @staticmethod
    def parse_yml(file_path: str):
        raise NotImplementedError
        try:
            import yaml
        except:
            print("can't import yaml")
            raise ImportError
        
    @staticmethod
    def sanity_check_configs(parsed_config) -> None:
        required = ('type', 'client', 'auth_request')
        for req in required: assert hasattr(parsed_config, req)
        assert (parsed_config.flask or parsed_config.django)
        
        assert ClientType.has_key(parsed_config.type.client_type), "'client_type' must be non-empty string"
        assert AuthorityType.has_key(parsed_config.type.authority_type), "'authority_type' must be non-empty string"
        assert parsed_config.type.framework == 'FLASK' or parsed_config.type.framework == 'DJANGO', "only FLASK & DJANGO supported right now"

        assert str(parsed_config.client.client_id), "'client_id' must be non-empty string"
        assert str(parsed_config.client.authority), "'authority' must be non-empty string"

        required = ['redirect_uri', 'scopes', 'response_type']
        for req in required: assert hasattr(parsed_config.auth_request, req)

        # if ClientType(parsed_config.type.client_type) is ClientType.CONFIDENTIAL:
        #     assert parsed_config.client.client_credential, (
        #     "'client_credential' must be non-empty string if "
        #     "'client_type' is ClientType.CONFIDENTIAL")

        if AuthorityType(parsed_config.type.authority_type) is AuthorityType.B2C:
            assert parsed_config.b2c, (
                "config must contain 'b2c' section if 'authority_type' is AuthorityType.B2C")
            
            # assert b2c has required keys:            
            required_keys = ['susi','password', 'profile']
            for key in required_keys:
                assert getattr(parsed_config.b2c, key).startswith('/'), (
                    f"`{key}` value under b2c must be non-empty string if "
                    "'authority_type' is AuthorityType.B2C")
        else:
            setattr(parsed_config, 'b2c', None)

        if parsed_config.type.framework == 'FLASK':
            assert parsed_config.flask.id_web_configs
            required_keys = ['prefix', 'sign_in', 'edit_profile', 'redirect', 'sign_out', 'post_sign_out']
            for key in required_keys:
                assert getattr(parsed_config.flask.auth_endpoints, key).startswith('/'), (
                    f"The `{key}` value under 'flask.auth_endpoints must be string starting with / if "
                    "'framework' is FLASK")
        elif parsed_config.type.framework == 'DJANGO':
            assert parsed_config.django.id_web_configs
            required_keys = ['prefix', 'sign_in', 'edit_profile', 'redirect', 'sign_out', 'post_sign_out']
            for key in required_keys:
                assert getattr(parsed_config.django.auth_endpoints, key), (f"The `{key}` value under 'django.auth_endpoints'"
                "must be non-empty string if 'framework' is DJANGO")


class AuthError(Exception):
    # basic auth exception
    pass
class AuthSecurityError(AuthError):
    # security check failed, abort auth attempt
    code = 400
    status = 400
    description = "security check failed (state or nonce)"
class OtherAuthError(AuthError):
    # unknown aad error, abort auth attempt
    code = 500
    status = 500
    description = "unknown error"
class TokenExchangeError(AuthError):
    # unknown aad error, abort auth attempt
    code = 500
    status = 500
    description = "failed to exchange auth code for token(s)"
class B2CPasswordError(AuthError):
    # login interrupted, must do password reset
    code = 300
    status = 300
    description = "password reset/redirect"

try:
    from werkzeug.exceptions import HTTPException
    from flask import request
    class NotAuthenticatedError(HTTPException, AuthError):
        """Flask HTTPException Error + IdWebPy AuthError: User is not authenticated."""
        code = 401
        status = 401
        description = 'User is not authenticated'
except:
    class NotAuthenticatedError(AuthError):
        """IdWebPy AuthError: User is not authenticated."""
        code = 401
        status = 401
        description = 'User is not authenticated'


try:
    from django.urls import path
    from django.shortcuts import redirect
    from django.urls import reverse
except:
    pass
import logging

class MsalViews:
    def __init__(self, ms_identity_web):
        self.logger = logging.getLogger('MsalViewsLogger')
        self.ms_identity_web = ms_identity_web
        self.prefix = self.ms_identity_web.aad_config.django.auth_endpoints.prefix + "/"
        self.endpoints = self.ms_identity_web.aad_config.django.auth_endpoints

    def url_patterns(self):
        return [
            path(self.endpoints.sign_in, self.sign_in, name=self.endpoints.sign_in),
            path(self.endpoints.edit_profile, self.edit_profile, name=self.endpoints.edit_profile),
            path(self.endpoints.redirect, self.aad_redirect, name=self.endpoints.redirect),
            path(self.endpoints.sign_out, self.sign_out, name=self.endpoints.sign_out),
            path(self.endpoints.post_sign_out, self.post_sign_out, name=self.endpoints.post_sign_out),
        ]

    def sign_in(self, request):
        self.logger.debug(f"{self.prefix}{self.endpoints.sign_in}: request received. will redirect browser to login")
        auth_url = self.ms_identity_web.get_auth_url(redirect_uri=request.build_absolute_uri(reverse(self.endpoints.redirect)))
        return redirect(auth_url)

    def edit_profile(self, request):
        self.logger.debug(f"{self.prefix}{self.endpoints.edit_profile}: request received. will redirect browser to edit profile")
        auth_url = self.ms_identity_web.get_auth_url(
                redirect_uri=request.build_absolute_uri(reverse(self.endpoints.redirect)),
                b2c_policy=self.ms_identity_web.aad_config.b2c.profile)
        return redirect(auth_url)

    def aad_redirect(self, request):
        self.logger.debug(f"{self.prefix}{self.endpoints.redirect}: request received. will process params")
        return self.ms_identity_web.process_auth_redirect(
            redirect_uri=request.build_absolute_uri(reverse(self.endpoints.redirect)),
            afterwards_go_to_url=reverse('index'))

    def sign_out(self, request):
        self.logger.debug(f"{self.prefix}{self.endpoints.sign_out}: signing out username: {request.identity_context_data.username}")
        return self.ms_identity_web.sign_out(request.build_absolute_uri(reverse(self.endpoints.post_sign_out)))    # send the user to Azure AD logout endpoint

    def post_sign_out(self, request):
        self.logger.debug(f"{self.prefix}{self.endpoints.post_sign_out}: clearing session for username: {request.identity_context_data.username}")
        self.ms_identity_web.remove_user(request.identity_context_data.username)  # remove user auth from session on successful logout
        return redirect(reverse('index'))           


from msal import SerializableTokenCache
import json

# TODO: make this a @dataclass ?

class IdentityContextData(object):
    SESSION_KEY='identity_context_data' #TODO: make configurable

    def __init__(self) -> None:
        self.clear()
        self.has_changed = False

    def clear(self) -> None:
        self._authenticated = False
        self._username = "anonymous"
        self._token_cache = None
        self._nonce = None
        self._state = None
        self._id_token_claims = {} # does this belong here? yes, Token/claims customization. TODO: if it does, add getter/setter, # ID tokens aren't cached so store this here? 
        self._access_token = None
        self._last_used_b2c_policy = []
        self._post_sign_in_url = None
        self.has_changed = True

    @property
    def authenticated(self) -> bool:
        return self._authenticated

    @authenticated.setter
    def authenticated(self, value: bool) -> None:
        self._authenticated = value
        self.has_changed = True

    @property
    def username(self) -> str:
        return self._username

    @username.setter
    def username(self, value: str) -> None:
        self._username = value
        self.has_changed = True

    @property
    def token_cache(self) -> str:
        cache = SerializableTokenCache()
        if self._token_cache:
            cache.deserialize(self._token_cache)
        return cache

    @token_cache.setter
    def token_cache(self, value: SerializableTokenCache) -> None:
        if value.has_state_changed:
            self._token_cache = value.serialize()
            self.has_changed = True

    @property
    def state(self) -> str:
        return self._state

    @state.setter
    def state(self, value: str) -> None:
        self._state = value
        self.has_changed = True

    @property
    def nonce(self) -> str:
        return self._nonce

    @nonce.setter
    def nonce(self, value: str) -> None:
        self._nonce = value
        self.has_changed = True

    # TODO: talk to MSIDWEB team
    # or browse the code about how to implement the following:
    @property
    def last_used_b2c_policy(self) -> str:
        if len(self._last_used_b2c_policy):
            return self._last_used_b2c_policy.pop()
        return None

    @last_used_b2c_policy.setter
    def last_used_b2c_policy(self, value: str) -> None:
        self._last_used_b2c_policy = [value]
        self.has_changed = True

    @property
    def post_sign_in_url(self) -> str:
        return self._post_sign_in_url

    @post_sign_in_url.setter
    def post_sign_in_url(self, value: str) -> None:
        self._post_sign_in_url = value
        self.has_changed = True


from setuptools import setup

setup(
    name='jabutiutils',
    description=u'Jabuti Technologies - Utils',
    version='0.0.4',
    license='MIT License',
    author='Thales Araújo',
    install_requires=['msal>=1.6.0,<2', 'django'],
)


from enum import Enum

# class Policy(Enum):
#     def __str__(self):
#         return str(self.value)
#     @staticmethod
#     def set_enum_values(config: dict):
#         Policy.SIGN_UP_SIGN_IN = config['susi']
#         Policy.PASSWORD_RESET = config[str(Policy.PASSWORD_KEY)]
#         Policy.EDIT_PROFILE = config[str(Policy.PROFILE_KEY)]
#     SUSI_KEY = 'susi'
#     PASSWORD_KEY = 'password'
#     PROFILE_KEY = 'profile'
#     NONE = ''

### AZURE AD AUTH OPTIONS ###
class ResponseType(Enum):
    def __str__(self):
        return str(self.value)
    PARAM_KEY = 'response_type'
    CODE = 'code'  # this is the default ResponseType used by MSAL Python
    TOKEN = 'token'
    ID_TOKEN = 'id_token'
    ID_TOKEN_TOKEN = 'id_token token'
    CODE_TOKEN = 'code token'
    CODE_ID_TOKEN_TOKEN = 'code id_token token'
    NONE = 'none'

class ResponseMode(Enum):
    def __str__(self):
        return str(self.value)
    QUERY = 'query' # this is the default ResponseMode for ResponseType.CODE
    FRAGMENT = 'fragment'
    FORM_POST = 'form_post' 


class RequestParameter(Enum):
    def __str__(self):
        return str(self.value)
    RESPONSE_TYPE = 'response_type'
    PROMPT = 'prompt'
    REDIRECT_URI = 'redirect_uri'
    STATE = 'state'
    NONCE = 'nonce'
    SCOPE = 'scope'
    CLIENT_ID = 'client_id'


class Prompt(Enum):
    def __str__(self):
        return str(self.value) 
    PARAM_KEY = 'prompt'
    LOGIN = 'login' # causes user to re-enter credentials even if logged in already - negates sso
    NONE = 'none'   # opposite of prompt=login - no prompt displayed if user is already logged in
    SELECT_ACCOUNT = 'select_account' # user must select their account from picker
    CONSENT = 'consent' # user is asked for consent, even if they have given it previously

### Client Type ###
class ClientType(Enum):
    def __str__(self):
        return str(self.value)
    @classmethod
    def has_key(cls, name):
        return name in cls.__members__
    CONFIDENTIAL = 'CONFIDENTIAL'
    PUBLIC = 'B2C'
    

### Client Tenant Type ###
class AuthorityType(Enum):
    def __str__(self):
        return str(self.value)
    @classmethod
    def has_key(cls, name):
        return name in cls.__members__
    SINGLE_TENANT = 'SINGLE_TENANT'
    MULTI_TENANT = 'MULTI_TENANT'
    B2C = 'B2C'
    

### AZURE ACTIVE DIRECTORY ERROR HANDLING CONSTANTS ###
class AADErrorResponse(Enum):
    def __str__(self):
        return str(self.value)
    #AAD B2C forgot password error code, found in description of redirect request:
    B2C_FORGOT_PASSWORD_ERROR_CODE='AADB2C90118'
    #The parameter under which the error codes are found (in requests to redirect endpoint):
    ERROR_CODE_PARAM_KEY='error_description'
    #The parameter that indicates error:
    ERROR_PARAM_KEY='error'


### AZURE ACTIVE DIRECTORY SIGN-OUT ###
class SignOut(Enum):
    # The AAD endpoint to log your user out
    ENDPOINT = '/oauth2/v2.0/logout'
    # post-logout param key that tells AAD to redirect the user back to the app
    REDIRECT_PARAM_KEY = f'post_logout_redirect_uri'